"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.savePage = exports.fetchAsset = exports.fixBaseURI = exports.fetchMedia = exports.loadPage = exports.getPage = exports.getBrowser = void 0;
const tslib_1 = require("tslib");
const lodash_memoize_1 = tslib_1.__importDefault(require("lodash.memoize"));
const puppeteer_core_1 = tslib_1.__importDefault(require("puppeteer-core"));
const jsdom_1 = require("jsdom");
const node_fetch_1 = tslib_1.__importDefault(require("node-fetch"));
const path_1 = require("path");
const node_toolkit_1 = require("@tech_query/node-toolkit");
const yaml_1 = require("yaml");
const fs_extra_1 = require("fs-extra");
const config_1 = require("./config");
const parser_1 = require("./parser");
exports.getBrowser = (0, lodash_memoize_1.default)(() => puppeteer_core_1.default.launch({
    executablePath: config_1.executablePath,
    args: [
        '--single-process',
        '--no-zygote',
        '--disable-dev-shm-usage',
        '--disable-gpu',
        '--no-sandbox',
        '--disable-setuid-sandbox',
        '--no-default-browser-check',
        '--disable-default-apps',
        '--disable-extensions',
        '--no-startup-window',
        '--no-first-run',
        '--disable-infobars',
        '--disable-hang-monitor',
        '--disable-popup-blocking',
        '--disable-prompt-on-repost',
        '--disable-sync',
        '--disable-translate',
        '--disable-bundled-ppapi-flash',
        '--safebrowsing-disable-auto-update',
        '--disable-component-update',
        '--ignore-certificate-errors',
        '--disable-client-side-phishing-detection',
        '--disable-logging',
        '--mute-audio'
    ]
}));
async function getPage() {
    const browser = await (0, exports.getBrowser)();
    const page = await browser.newPage();
    await Promise.all([
        page.setRequestInterception(true),
        page.setUserAgent(config_1.userAgent)
    ]);
    page.setDefaultNavigationTimeout(0);
    page.on('request', request => {
        if (['image', 'media'].includes(request.resourceType()))
            request.abort();
        else
            request.continue();
    });
    return page;
}
exports.getPage = getPage;
async function loadPage(URI, root_selector = 'body') {
    if (!config_1.executablePath)
        return jsdom_1.JSDOM.fromURL(URI, {
            userAgent: config_1.userAgent,
            pretendToBeVisual: true,
            runScripts: 'outside-only'
        });
    const page = await getPage();
    await page.goto(URI);
    await page.waitForSelector(`${root_selector}:not(:empty)`);
    const HTML = await page.content();
    await page.close();
    return new jsdom_1.JSDOM(HTML, { url: URI });
}
exports.loadPage = loadPage;
function fetchMedia(root, root_path = '') {
    return tslib_1.__asyncGenerator(this, arguments, function* fetchMedia_1() {
        for (const media of root.querySelectorAll(parser_1.MediaSelector)) {
            try {
                var { pathname, protocol, href } = (0, parser_1.sourcePathOf)(media);
            }
            catch (_a) {
                media.remove();
            }
            if (protocol === 'data:')
                var { MIME, data } = yield tslib_1.__await((0, node_toolkit_1.blobFrom)(href));
            else {
                const response = yield tslib_1.__await((0, node_fetch_1.default)(href, {
                    headers: { 'User-Agent': config_1.userAgent }
                }));
                var [MIME] = response.headers.get('Content-Type').split(';'), data = yield tslib_1.__await(response.buffer());
            }
            const name = yield tslib_1.__await((0, parser_1.createFilePath)(data, pathname, root_path));
            media.src = name;
            if (media.style.display === 'none')
                media.style.display = '';
            yield yield tslib_1.__await({ MIME, name, data });
        }
    });
}
exports.fetchMedia = fetchMedia;
function fixBaseURI(document, baseURI) {
    const base = document.createElement('base');
    base.href = baseURI;
    document.head.append(base);
    for (const media of document.querySelectorAll(parser_1.MediaSelector))
        media.setAttribute('src', media.src);
    for (const link of document.querySelectorAll(parser_1.LinkSelector))
        link.setAttribute('href', link.href);
    base.remove();
}
exports.fixBaseURI = fixBaseURI;
function fetchAsset(document, { scope, rootSelector, baseURI = '', markdown }, meta) {
    return tslib_1.__asyncGenerator(this, arguments, function* fetchAsset_1() {
        var e_1, _a;
        const root = (0, parser_1.parseContent)(document, rootSelector);
        try {
            for (var _b = tslib_1.__asyncValues(fetchMedia(root, scope)), _c; _c = yield tslib_1.__await(_b.next()), !_c.done;) {
                const media = _c.value;
                yield yield tslib_1.__await(media);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) yield tslib_1.__await(_a.call(_b));
            }
            finally { if (e_1) throw e_1.error; }
        }
        if (baseURI)
            fixBaseURI(document, baseURI);
        let markup = `<meta charset="UTF-8">
${root.innerHTML.trim().replace(/(\n\s*)+/g, '\n')}`;
        if (markdown)
            markup = `---
${(0, yaml_1.stringify)(meta)}
---

${parser_1.convertor.turndown(markup)}`;
        yield yield tslib_1.__await({
            MIME: `text/${markdown ? 'markdown' : 'html'}`,
            name: `${(0, parser_1.fileNameOf)(scope)}.${markdown ? 'md' : 'html'}`,
            data: Buffer.from(markup)
        });
    });
}
exports.fetchAsset = fetchAsset;
async function savePage({ source, rootSelector, baseURI = '', markdown, rootFolder = process.cwd() }) {
    var e_2, _a;
    console.time('Fetch');
    const scope = (0, path_1.parse)(source).name, { window: { document } } = await loadPage(source);
    const meta = (0, parser_1.parseMeta)(document, config_1.meta_tag);
    const root_path = (0, path_1.join)(rootFolder, ...(meta.categories || []).map(parser_1.fileNameOf));
    try {
        for (var _b = tslib_1.__asyncValues(fetchAsset(document, { scope, rootSelector, baseURI, markdown }, meta)), _c; _c = await _b.next(), !_c.done;) {
            const { name, data } = _c.value;
            const filePath = (0, path_1.join)(root_path, name);
            await (0, fs_extra_1.outputFile)(filePath, data);
            console.log('[save] ' + filePath);
        }
    }
    catch (e_2_1) { e_2 = { error: e_2_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b.return)) await _a.call(_b);
        }
        finally { if (e_2) throw e_2.error; }
    }
    console.log('--------------------');
    console.timeEnd('Fetch');
}
exports.savePage = savePage;
